package lsedit;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;


class SetNewGridDialog extends JDialog implements ChangeListener, ActionListener
{
	class GridImage extends JComponent
	{
		public GridImage()
		{
			super();
			setPreferredSize(new Dimension(400, 100));
		}

		public void paintComponent(Graphics g)
		{
			int	grid = m_gridSize;
			int	w    = getWidth();
			int h    = getHeight();
			int	i;

			g.drawRect(0,0, w-1, h-1);

			if (grid > 1) {
				Color color   = m_gridColor;
				Color inverse = ColorCache.getInverse(color.getRGB());

				g.setColor(inverse);
				g.fillRect(1,1,w-2,w-2);

				g.setColor(color);
				for (i = grid; i < h; i += grid) {
					g.drawLine(0,i,w-1,i);
				}
				for (i = grid; i < w; i += grid) {
					g.drawLine(i,0,i, h-1);
		}	}	}
	}

	JFrame		m_frame;
	JSlider		m_gridSlider;
	GridImage	m_gridImage;
	JLabel		m_feedback;
	JButton		m_color;
	int			m_gridSize;
	Color		m_gridColor;

	public SetNewGridDialog(JFrame frame, int gridSize, Color gridColor)
	{
		super(frame, "Set grid size", true);

		Container	contentPane = getContentPane();
		Font		font, bold;

		m_frame      = frame;
		m_gridSize   = gridSize;
		m_gridColor  = gridColor;
		font         = FontCache.getDialogFont();
		bold         = font.deriveFont(Font.BOLD);

		JLabel sliderLabel = new JLabel("Grid size when moving entities", JLabel.CENTER);
		sliderLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		sliderLabel.setFont(bold);

		JLabel feedback  = new JLabel("Current grid size " + gridSize, JLabel.CENTER);
		feedback.setAlignmentX(Component.CENTER_ALIGNMENT);
		feedback.setFont(font);
		feedback.setForeground(Color.BLUE);
		m_feedback = feedback;

		m_gridImage = new GridImage();

		JSlider gridSlider = new JSlider(JSlider.HORIZONTAL, 0, 100, m_gridSize);

		m_gridSlider = gridSlider;

		gridSlider.addChangeListener(this);

		gridSlider.setMajorTickSpacing(10);
		gridSlider.setMinorTickSpacing(2);
		gridSlider.setPaintTicks(true);
		gridSlider.setPaintLabels(true);
		gridSlider.setFont(font);
		gridSlider.setPreferredSize(new Dimension(400, 50));

		gridSlider.setBorder(BorderFactory.createEmptyBorder(0,0,10,0));

		JPanel	panel = new JPanel();
		panel.setLayout(new BorderLayout());
		panel.add(feedback, BorderLayout.NORTH);
		panel.add(m_gridImage,BorderLayout.CENTER);

		m_color = new JButton("Colour");
		m_color.addActionListener(this);
		panel.add(m_color, BorderLayout.SOUTH);

		contentPane.add(sliderLabel, BorderLayout.NORTH);
		contentPane.add(gridSlider,  BorderLayout.CENTER);
		contentPane.add(panel,       BorderLayout.SOUTH);

		if (frame != null) {
			setLocation(frame.getX()+200, frame.getY()+300);
		}
		pack();
		setVisible(true);
	}
	
	public int getGridSize()
	{
		return m_gridSize;
	}
	
	public Color getGridColor()
	{
		return m_gridColor;
	}

	// ActionListener interface

	public void actionPerformed(ActionEvent ev)
	{
		Object source = ev.getSource();

		if (source == m_color) {
			ColorChooser colorChooser = new ColorChooser(m_frame, "Pick a grid color", m_gridColor, true /* include alpha */, false /* Don't allow null */);
			m_gridColor = colorChooser.getColor();
			colorChooser.dispose();
			m_gridImage.repaint();
	}	}

	public void stateChanged(ChangeEvent ev)
	{
		Object source = ev.getSource();

		if (source == m_gridSlider) {
			int value = (int) m_gridSlider.getValue();

			if (value < 1) {
				value = 1;
			}
			m_gridSize = value;
			m_gridImage.repaint();
	}	}
}
	
public class OptionsDialog extends JDialog implements ActionListener { 

	protected LandscapeEditorCore	m_ls;

	// JPanel1 stuff
	
	private static final int LOAD_RULE = 0;

	protected final static String[] m_combo_titles1 = 
	{
		"Load TA options:"
	};

	public final static String[]	g_loadchoices = new String[]
	{
		"Load",				// Fixed position
		"Ignore",			// Fixed position
		"Prompt",
	};
	protected JComboBox				m_loadchoices = new JComboBox(g_loadchoices);

	protected final static String[] m_checkbox_titles1 = 
	{
		 "Show description box:",
		 "Show feedback box:",
		 "Show Tab Box on left:",
		 "Scroll through tabs:",
		 "Always show scrollbars:",
		 "Sort the Table of Contents:",

		 "Show clients at top:",
		 "Show clients:",
		 "Show suppliers:",
		 "Use client/supplier compaction:",
		 "Client/suppliers use visible edges:",
		 "Client/suppliers use visible entities:",
		 
		 "Lift edges under closed entities:",
		 "Show destination cardinal counts:",
		 "Show source cardinal counts:",

		 "Group the queried items:",
		 "Queries persist:",
		 "Nearest ancestor changing hierarchy:",
		 "Hide empty classes in the legend:",
		 "Show member counts in the legend:",
		 "Show inheritance in legend:",
		 "Show legend labels in black:",

		 "Show grid:",
		 "Snap entities to grid:"
	};
	
	public final static String[]	g_optionchoices = new String[]
								{
									"Reset",		// Fixed position
									"Default",		// Fixed position
									"Landscape",
									"Current",
								};
						
	protected JComboBox	m_optionchoices = new JComboBox(g_optionchoices);
	
	protected JComboBox[]			m_comboboxes1;
	protected JCheckBox[]			m_checkboxes1;
	protected JButton				m_setGridSize;

	// JPanel2 stuff
	
	protected final static String[] m_textfield_titles2 = 
							{
								"Width of edge:",
								"Length of arrow edge:",
								"Width of arrow arc:",
								"3D highlighting weight:",
								"Incidental shadow size:",
								"Angle of label:",
								"Entity hover factor:",
								"Horizontal zoom factor:",
								"Vertical zoom factor:",
								"Icon search path:"
							};
							
	protected final static String[] m_textfield_units2 = 
							{
								"Pixels",
								"Pixels",
								"Radians",
								"Pixels",
								"Pixels",
								"Degrees",
								"Scale",
								"Scale",
								"Scale",
								""
							};

	private static final int EDGE_MODE = 0;
	private static final int ICON_RULE = 1;


	protected final static String[] m_combo_titles2 = 
							{
								"Edge mode:",
								"Show icons:"
							};

		public final static String[]	g_edgechoices = new String[]
	{
		"Inflection",		// Fixed position
		"Top/Bottom",		// Fixed position
		"Direct",
		"Direct+Side"
	};
	
	protected JComboBox				m_edgechoices = new JComboBox(g_edgechoices);

	protected final static String[] m_combo_units2 = 
							{
								"",
								""
							};

	public final static String[]	g_iconchoices = new String[]
								{
									"No icons",			// Fixed position
									"Plain icons",		// Fixed position
									"Labels centered",
									"Labels bottom",
									"Labels top",
									"Empty cache"
								};	
	
	protected JComboBox	m_iconchoices = new JComboBox(g_iconchoices);

	protected final static String[] m_checkbox_titles2 = 
							{
								 "Fixed Icon Shape:",
								 "Fill arrow head:",
								 "Center arrow head:",
								 "Permanently weight arrow head:",
								 "3D in Black&White:",
								 "Show fixed edge labels:",
								 "Edge labels angle w.r.t edge:",
								 "Show edge tooltip:",
								 "Variable arrow color:",
								 "Invert edge label background:",
								 "Label color inverts background",
								 "Invert entity label background",
 								 "Make entity label black/white"
							};
	
	protected JTextField[]			m_textfields2;
	protected JComboBox[]			m_comboboxes2;
	protected JCheckBox[]			m_checkboxes2;
	
			
	protected Container				m_contentPane;
	protected JComponent			m_centerThing;
	protected JScrollPane			m_scrollPane1;
	protected JScrollPane			m_scrollPane2;
	protected boolean				m_use_tabs = false;

	// Other stuff
	
	protected Option	m_current       = new Option();
	
	static protected final int BUTTON_OK      = 0;
	static protected final int BUTTON_CANCEL  = 1;
	static protected final int BUTTON_LOAD    = 2;
	static protected final int BUTTON_SAVE    = 3;
	static protected final int BUTTON_TAB     = 4;
	static protected final int BUTTON_HELP    = 5;

	protected final static String[] m_button_titles =
							{
								"Ok",
								"Cancel",
								"Load",
								"Save",
								"Tab",
								"Help"
							};

	protected final static String[] m_button_tips =
							{
								"Set values shown as current options",
								"Change nothing",
								"Load values show with specified options",
								"Set specified options to values shown",
								"Switch to/from tabbed view",
								"Bring up browser describing options"
							};

	protected JLabel				m_message;
	protected JButton[]				m_buttons;
	
	protected Option getTarget()
	{
		switch (m_optionchoices.getSelectedIndex()) {
		case 1:
			return Options.getDefaultOptions();
		case 2:
			return Options.getLandscapeOptions();
		case 3:
			return Options.getDiagramOptions();
		}
		return(null);
	}
	
	protected String getTargetName()
	{
		int choice = m_optionchoices.getSelectedIndex();
		
		if (choice >= 0 && choice < g_optionchoices.length) {
			return(g_optionchoices[choice]);
		}
		return "?";
	}
	
	protected void loadValues1()
	{
		Option			option     = m_current;
		JComboBox[]		comboboxes = m_comboboxes1;
		JCheckBox[]		checkboxes = m_checkboxes1;
		int				ival       = -1;
		boolean			bool       = false;
		int				i, lth;
	
		lth = m_combo_titles1.length;
		for (i = 0; i < lth; ++i) {
			switch (i) {
			case LOAD_RULE:
				ival     = option.getLoadMode();
				break;
			}
			comboboxes[i].setSelectedIndex(ival);
		}
		
		lth = m_checkbox_titles1.length;
		for (i = 0; i < lth; ++i) {
			bool = option.getMainBooleanParameter(i);
			checkboxes[i].setSelected(bool);
		}
		
		m_setGridSize.setText("" + option.getGridSize());
		m_setGridSize.setForeground(option.getGridColor());
	}
	
	protected void loadValues2()
	{
		Option			option     = m_current;
		JTextField[]	textfields = m_textfields2;
		JComboBox[]		comboboxes = m_comboboxes2;
		JCheckBox[]		checkboxes = m_checkboxes2;
		String			value      = null;
		int				ival       = -1;
		boolean			bool       = false;
		int				i, lth;
	
		lth = m_textfield_titles2.length;
		for (i = 0; i < lth; ++i) {
			value = option.getArrowNumericParameter(i);
			textfields[i].setText(value);
		}
		
		lth = m_combo_titles2.length;
		for (i = 0; i < lth; ++i) {
			switch (i) {
			case EDGE_MODE:
				ival     = option.getEdgeMode();
				break;
			case ICON_RULE:
				ival     = option.getIconRule();
				break;
			}
			comboboxes[i].setSelectedIndex(ival);
		}
		
		lth = m_checkbox_titles2.length;
		for (i = 0; i < lth; ++i) {
			bool = option.getArrowBooleanParameter(i);
			checkboxes[i].setSelected(bool);
		}
	}
	
	protected void saveValues1()
	{
		Option			option     = m_current;
		JComboBox[]		comboboxes = m_comboboxes1;
		JCheckBox[]		checkboxes = m_checkboxes1;
		int				i, lth;
	
		lth = m_combo_titles1.length;
		for (i = 0; i < lth; ++i) {
			switch (i) {
			case LOAD_RULE:
				option.setLoadMode(comboboxes[i].getSelectedIndex());
				break;
		}	}
		
		lth = m_checkbox_titles1.length;
		for (i = 0; i < lth; ++i) {
			option.setMainBooleanParameter(i, checkboxes[i].isSelected());
		}
	}
	
	protected String saveValues2()
	{
		Option			option     = m_current;
		JTextField[]	textfields = m_textfields2;
		JComboBox[]		comboboxes = m_comboboxes2;
		JCheckBox[]		checkboxes = m_checkboxes2;
		int				i, index, lth;
		String			ret, ret1;
	
		ret = null;
		lth = m_textfield_titles2.length;
		for (i = 0; i < lth; ++i) {
			ret1 = option.setArrowNumericParameter(i, textfields[i].getText());
			if (ret1 != null && ret == null) {
				ret = ret1;
		}	}
		
		lth = m_combo_titles2.length;
		for (i = 0; i < lth; ++i) {
			index = comboboxes[i].getSelectedIndex();
			switch (i) {
			case EDGE_MODE:
				option.setEdgeMode(index);
				break;
			case ICON_RULE:
				if (index == Option.ICON_RULE_EMPTY) {
					IconCache.clear();
					index = Option.ICON_RULE_NONE;
				} 
				option.setIconRule(index);
				break;
		}	}
		
		lth = m_checkbox_titles2.length;
		for (i = 0; i < lth; ++i) {
			option.setArrowBooleanParameter(i, checkboxes[i].isSelected());
		}
		return ret;
	}
	
	protected void addCenterThing()
	{
		if (m_use_tabs) {
			JTabbedPane tabbedPane = new JTabbedPane();
				
			tabbedPane.add("Main", m_scrollPane1);
			tabbedPane.add("Visualisation", m_scrollPane2);	
			m_centerThing = tabbedPane;
		} else {
			JPanel	topPanel = new JPanel();
			topPanel.setLayout( new BorderLayout() );
 			topPanel.add(BorderLayout.WEST, m_scrollPane1);
			topPanel.add(BorderLayout.EAST, m_scrollPane2);
	        m_centerThing = topPanel;
		}
		m_contentPane.add( BorderLayout.CENTER, m_centerThing );
	}

	protected OptionsDialog(LandscapeEditorCore ls) //Constructor
	{
		super(ls.getFrame(), "Modify LSEdit options", true); //false if non-modal

		Container	contentPane;
		JLabel		label;
		Font		font, bold;
		JTextField	textfield;
		JCheckBox	checkbox;
		JButton		button;
		int			i;
		String		tip;
		JComboBox	combobox;

		m_ls         = ls;
		font         = FontCache.getDialogFont();
		bold         = font.deriveFont(Font.BOLD);

		m_contentPane = contentPane = getContentPane();
        contentPane.setLayout( new BorderLayout() );

		setForeground(ColorCache.get(0,0,0));
		setBackground(ColorCache.get(192,192,192));

		setFont(font);

		// Top panel1 logic
		
		JPanel topPanel1    = new JPanel();
		JPanel leftPanel1   = new JPanel();
		JPanel centerPanel1 = new JPanel();

		GridLayout gridLayout;
		
		topPanel1.setLayout( new BorderLayout() );
		gridLayout = new GridLayout(26,1);
		gridLayout.setVgap(0);
		leftPanel1.setLayout( gridLayout);

		gridLayout = new GridLayout(26,1);
		gridLayout.setVgap(0);
		centerPanel1.setLayout( gridLayout);

		topPanel1.add( BorderLayout.WEST, leftPanel1);
		topPanel1.add( BorderLayout.EAST, centerPanel1);

		// ----------

		m_comboboxes1 = new JComboBox[m_combo_titles1.length];
		for (i = 0; i < m_combo_titles1.length; ++i) {
			label = new JLabel(m_combo_titles1[i], JLabel.RIGHT);
			label.setFont(bold);
			leftPanel1.add(label);
			
			switch (i) {
			case LOAD_RULE:
				combobox = m_loadchoices;
				break;
			default:
				combobox = null;
			}
			m_comboboxes1[i] = combobox;
			combobox.setFont(font);
			centerPanel1.add(combobox);
		}

		m_checkboxes1 = new JCheckBox[m_checkbox_titles1.length];

		for (i = 0; i < m_checkbox_titles1.length; ++i) {
			label = new JLabel(m_checkbox_titles1[i], JLabel.RIGHT);
			label.setFont(bold);
			leftPanel1.add(label);
			m_checkboxes1[i] = checkbox = new JCheckBox("");
			checkbox.setFont(font);
			centerPanel1.add(checkbox);
		}

		label = new JLabel("Set Diagram Grid:", JLabel.RIGHT);
		label.setFont(bold);
		leftPanel1.add(label);
		m_setGridSize = new JButton();
		centerPanel1.add(m_setGridSize);
		m_setGridSize.addActionListener(this);

		m_scrollPane1 = new JScrollPane(topPanel1);
		
		// JPanel2 setup
		
		JPanel topPanel2    = new JPanel();
		JPanel leftPanel2   = new JPanel();
		JPanel centerPanel2 = new JPanel();
		JPanel rightPanel2  = new JPanel();

		topPanel2.setLayout( new BorderLayout() );
		gridLayout = new GridLayout(25,1);
		gridLayout.setVgap(0);
		leftPanel2.setLayout( gridLayout);

		gridLayout = new GridLayout(25,1);
		gridLayout.setVgap(0);
		centerPanel2.setLayout( gridLayout);

		gridLayout = new GridLayout(25,1);
		gridLayout.setVgap(0);
		rightPanel2.setLayout( gridLayout);

		topPanel2.add( BorderLayout.WEST, leftPanel2);
		topPanel2.add( BorderLayout.EAST, rightPanel2);
		topPanel2.add( BorderLayout.CENTER, centerPanel2);

		// ----------

		m_textfields2 = new JTextField[Option.m_arrow_numeric_tags.length];
		for (i = 0; i < m_textfield_titles2.length; ++i) {
			label = new JLabel(m_textfield_titles2[i], JLabel.RIGHT);
			label.setFont(bold);
			leftPanel2.add(label);

			m_textfields2[i] = textfield = new JTextField(12);
			textfield.setFont(font);
			textfield.addActionListener(this);
			centerPanel2.add(textfield);
			
			label = new JLabel(m_textfield_units2[i], JLabel.LEFT);
			label.setFont(bold);
			rightPanel2.add(label);
		}

		m_comboboxes2 = new JComboBox[m_combo_titles2.length];
		for (i = 0; i < m_combo_titles2.length; ++i) {
			label = new JLabel(m_combo_titles2[i], JLabel.RIGHT);
			label.setFont(bold);
			leftPanel2.add(label);
			
			switch (i) {
			case EDGE_MODE:
				combobox = m_edgechoices;
				break;
			case ICON_RULE:
				combobox = m_iconchoices;
				break;
			default:
				combobox = null;
			}
			m_comboboxes2[i] = combobox;
			combobox.setFont(font);
			centerPanel2.add(combobox);
			
			label = new JLabel(m_combo_units2[i], JLabel.LEFT);
			label.setFont(bold);
			rightPanel2.add(label);
		}

		m_checkboxes2 = new JCheckBox[m_checkbox_titles2.length];

		for (i = 0; i < m_checkbox_titles2.length; ++i) {
			label = new JLabel(m_checkbox_titles2[i], JLabel.RIGHT);
			label.setFont(bold);
			leftPanel2.add(label);
			m_checkboxes2[i] = checkbox = new JCheckBox("");
			checkbox.setFont(font);
			centerPanel2.add(checkbox);
		}
		
		
		m_scrollPane2 = new JScrollPane(topPanel2);
	
		m_current.setMainOptionsTo(Options.getDiagramOptions());
		m_current.setArrowOptionsTo(Options.getDiagramOptions());
		
		// ------------

		m_message = new JLabel(" ", JLabel.CENTER);
		m_message.setFont(font);
		m_message.setForeground(Color.RED);

		m_message.setSize(400,50);
		contentPane.add( BorderLayout.NORTH, m_message);

		addCenterThing();

		// --------------
		// Use a FlowLayout to center the button and give it margins.

		JPanel bottomPanel = new JPanel();

		bottomPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 15, 15));

		m_optionchoices.setFont(bold);
		m_optionchoices.setSelectedIndex(3);
		m_optionchoices.addActionListener(this);
		
		loadValues1();
		loadValues2();
		
		m_loadchoices.setEnabled(false);
	
		bottomPanel.add(m_optionchoices);

		m_buttons = new JButton[m_button_titles.length];
		for (i = 0; i < m_button_titles.length; ++i) {
			m_buttons[i] = button = new JButton(m_button_titles[i]);
			button.setFont(bold);
			tip = m_button_tips[i];
			if (tip != null) {
				button.setToolTipText(tip);
			}
			button.addActionListener(this);
			bottomPanel.add(button);
		}

		contentPane.add( BorderLayout.SOUTH, bottomPanel);

		// Resize the window to the preferred size of its components
		setLocation(20, 20);
		pack();
		setVisible(true);
	}

	// ActionListener interface

	public void actionPerformed(ActionEvent ev)
	{
		Object			source;
		JTextField		textfield;
		String			msg, msg1;
		int				state, i;
		Option			target, current;
		boolean			firstPane, secondPane;

		// Pop down the window when the button is clicked.
		// System.out.println("event: " + ev);

		source = ev.getSource();
		
		if (source == m_optionchoices) {
			int index = m_optionchoices.getSelectedIndex();
			m_buttons[BUTTON_SAVE].setEnabled(index != 0);
			m_loadchoices.setEnabled(index != 3);
			return;
		}			

		current = m_current;

		if (source == m_setGridSize) {
			SetNewGridDialog	dialog    = new SetNewGridDialog(m_ls.getFrame(), current.getGridSize(), current.getGridColor());
			
			current.setGridSize(dialog.getGridSize());
			current.setGridColor(dialog.getGridColor());
			
			m_setGridSize.setText("" + current.getGridSize());
			m_setGridSize.setForeground(current.getGridColor());

			dialog.dispose();
			dialog = null;
			return;
		}

		state = -1;
		for (i = 0; i < m_button_titles.length; ++i) {
			if (source == m_buttons[i]) {
				state = i;
				break;
		}	}

		switch (state) {
		case BUTTON_CANCEL:
			break;
		case BUTTON_LOAD:
			target = getTarget();
			
			if (!m_use_tabs) {
				firstPane  = true;
				secondPane = true;
				msg        = "All"; 
			} else if (((JTabbedPane) m_centerThing).getSelectedComponent() == m_scrollPane1) {
				firstPane  = true;
				secondPane = false;
				msg        = "Main"; 
			} else {
				firstPane  = false;
				secondPane = true;
				msg        = "Visualisation"; 
			}
			if (firstPane) {
				current.setMainOptionsTo(target);
				loadValues1();
			}
			if (secondPane) {
				current.setArrowOptionsTo(target);
				loadValues2();
			}
			m_message.setText(msg + " options loaded from " + getTargetName() + " values");
			return;
		case BUTTON_SAVE:
		case BUTTON_OK:
			if (state == BUTTON_OK) {
				target = Options.getDiagramOptions();
			} else {
				target = getTarget();
			}
			if (target == null) {
				m_message.setText("Can't alter factory reset options");
				return;
			}
			if (!m_use_tabs) {
				firstPane  = true;
				secondPane = true;
				msg        = "All"; 
			} else if (((JTabbedPane) m_centerThing).getSelectedComponent() == m_scrollPane1) {
				firstPane  = true;
				secondPane = false;
				msg        = "Main"; 
			} else {
				firstPane  = false;
				secondPane = true;
				msg        = "Visualisation"; 
			}
		
			if (secondPane) {
				msg1 = saveValues2();
				if (msg1 != null) {
					m_message.setText(msg1);
					return;
				}
				target.setArrowOptionsTo(current);
			}
			if (firstPane) {
				saveValues1();
				target.setMainOptionsTo(current);
			}
			if (state == BUTTON_OK) {
				break;
			}
			m_message.setText(msg + " values saved to " + getTargetName() + " options");
			return;
		case BUTTON_TAB:
			m_use_tabs = !m_use_tabs;
			m_contentPane.remove(m_centerThing);
			addCenterThing();
			m_buttons[BUTTON_TAB].setText(m_use_tabs ? "Detab" : "Tab");
			pack();
			return;
		case BUTTON_HELP:
			m_ls.showURL(m_ls.m_helpURL + "/options.html", LsLink.TARGET_HELP);
		default:
			return;
		}
		
		setVisible(false);
		return;
	}

	public static void create(LandscapeEditorCore ls) 
	{
		Option option     = Options.getDiagramOptions();
		Option oldOptions = new Option();
		
		oldOptions.setArrowOptionsTo(option);
		oldOptions.setMainOptionsTo(option);

		OptionsDialog optionsDialog = new OptionsDialog(ls);

		optionsDialog.dispose();

		option.optionsChanged(ls, oldOptions);
	}
} 



